-- Stateful raycasting library

local rc = {}

local lg = love.graphics

-- view frustum's near and far clip; if far minus near is a power of two, some
-- calculations will be exact
local near, far = 0.0625, 1024.0625

-- Maximum number of vertices - preferably a multiple of 4
local vmax = 3000

local atypes = {
  {"VertexPosition", "float", 3},
  {"VertexTexCoord", "float", 2},
}

-- Position and rotation
rc.posrot = {1,0,0,0,
             0,1,0,0,
             0,0,1,0,
             0,0,0,1}

local oldvnum, vnum
local vmap = {}

local cos, sin = math.cos, math.sin

-- Multiply c = a * b (where a, b, c are 4x4 matrices in column-major order)
local function mul4x4(c, a, b)
  c[ 1] = a[ 1] * b[ 1] + a[ 5] * b[ 2] + a[ 9] * b[ 3] + a[13] * b[ 4]
  c[ 2] = a[ 2] * b[ 1] + a[ 6] * b[ 2] + a[10] * b[ 3] + a[14] * b[ 4]
  c[ 3] = a[ 3] * b[ 1] + a[ 7] * b[ 2] + a[11] * b[ 3] + a[15] * b[ 4]
  c[ 4] = a[ 4] * b[ 1] + a[ 8] * b[ 2] + a[12] * b[ 3] + a[16] * b[ 4]
  c[ 5] = a[ 1] * b[ 5] + a[ 5] * b[ 6] + a[ 9] * b[ 7] + a[13] * b[ 8]
  c[ 6] = a[ 2] * b[ 5] + a[ 6] * b[ 6] + a[10] * b[ 7] + a[14] * b[ 8]
  c[ 7] = a[ 3] * b[ 5] + a[ 7] * b[ 6] + a[11] * b[ 7] + a[15] * b[ 8]
  c[ 8] = a[ 4] * b[ 5] + a[ 8] * b[ 6] + a[12] * b[ 7] + a[16] * b[ 8]
  c[ 9] = a[ 1] * b[ 9] + a[ 5] * b[10] + a[ 9] * b[11] + a[13] * b[12]
  c[10] = a[ 2] * b[ 9] + a[ 6] * b[10] + a[10] * b[11] + a[14] * b[12]
  c[11] = a[ 3] * b[ 9] + a[ 7] * b[10] + a[11] * b[11] + a[15] * b[12]
  c[12] = a[ 4] * b[ 9] + a[ 8] * b[10] + a[12] * b[11] + a[16] * b[12]
  c[13] = a[ 1] * b[ 9] + a[ 5] * b[14] + a[ 9] * b[15] + a[13] * b[16]
  c[14] = a[ 2] * b[13] + a[ 6] * b[14] + a[10] * b[15] + a[14] * b[16]
  c[15] = a[ 3] * b[13] + a[ 7] * b[14] + a[11] * b[15] + a[15] * b[16]
  c[16] = a[ 4] * b[13] + a[ 8] * b[14] + a[12] * b[15] + a[16] * b[16]
end

local warnings = {}
local function send(shader, ...)
  local x, y = pcall(shader.send, shader, ...)
  if not x then
    if not warnings[y] then
      warnings[y] = true
      print("Error in Shader:send: " .. y)
    end
  end
end

-- Construct a projection matrix suitable for a 3D world view where +x is
-- forward, +y is left and +z is up. It includes the necessary rotations to
-- make the camera point where we desire, to save calculations. The 99's
-- are placeholders as they depend on the focal distance and width/height
-- and are calculated at runtime.
--
-- This matrix is constructed as proj * xrot * zrot, where proj is a
-- "standard" frustum projection matrix, and both xrot and zrot are 90°
-- rotations. xrot is used to compensate for the fact that the OpenGL
-- default view is a top view, therefore it rotates 90° CCW on X to point
-- north rather than down; zrot compensates for the fact that 0° is
-- assumed to point east, not north, therefore it's a 90° CW turn over Z.
-- Remember all matrices here are in column-major order, so this looks
-- transposed with respect to the actual matrix.
local proj = {0, 0, (far + near) / (far - near), 1,
              99, 0, 0, 0,
              0, 99, 0, 0,
              0, 0, (-2)*near*far / (far - near), 0}

function rc.init(tex, ground, ceil)
  rc.shader = lg.newShader [[
    extern mat4 proj;
    vec4 position(mat4 transform, vec4 vertex)
    {
      // ignore love2d's matrix and use our own
      return proj * vertex;
    }
  ]]
  lg.setShader(rc.shader)

  rc.mesh = lg.newMesh(atypes, vmax, "triangles", "dynamic")
  rc.mesh:setTexture(tex)

  rc.setCameraFocalDist(0.5)
  oldvnum, vnum = -1, 0
  -- clear vmap
  for i = #vmap, 1, -1 do vmap[i] = nil end
end

function rc.setCameraFocalDist(d)
  rc.fd = d
end

function rc.setCamera(posx, posy, posz, yaw)
  local c, s = cos(yaw), sin(yaw)
  rc.posrot[1] = c
  rc.posrot[2] = -s
  rc.posrot[5] = s
  rc.posrot[6] = c
  rc.posrot[13] = -(posx*c + posy*s)
  rc.posrot[14] =   posx*s - posy*c
  rc.posrot[15] = -posz
end

function rc.rect3d(x1, y1, z1, x2, y2, z2, x3, y3, z3, u1, v1, u2, v2, w, h)
  if vnum + 4 > vmax then
    error("rect3d maximum capacity exceeded")
  end
  vnum = vnum + 4
  local u, v = u1 / w, v1 / h
  rc.mesh:setVertex(vnum-3, x1, y1, z1, u, v)
  local vv = v
  v = v2 / h
  rc.mesh:setVertex(vnum-2, x2, y2, z2, u, v)
  u = u2 / w
  rc.mesh:setVertex(vnum-1, x3, y3, z3, u, vv)
  rc.mesh:setVertex(vnum, x2 + x3 - x1, y2 + y3 - y1, z2 + z3 - z1, u, v)

  vmap[(vnum-4)*1.5 + 1] = vnum - 3
  vmap[(vnum-4)*1.5 + 2] = vnum - 2
  vmap[(vnum-4)*1.5 + 3] = vnum - 1
  vmap[(vnum-4)*1.5 + 4] = vnum - 1
  vmap[(vnum-4)*1.5 + 5] = vnum - 2
  vmap[(vnum-4)*1.5 + 6] = vnum - 0
end

local result = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}

function rc.draw()
  local w, h = lg.getDimensions()

  -- Complete the projection matrix, now that we have w, h and focal distance
  proj[5] = -h / (w * rc.fd)
  proj[10] = 1 / rc.fd
  mul4x4(result, proj, rc.posrot)
  send(rc.shader, 'proj', result)

  if vnum ~= oldvnum then
    -- added rectangles
    rc.mesh:setDrawRange(1, #vmap)
    rc.mesh:setVertexMap(vmap)
    oldvnum = vnum
  end
  lg.draw(rc.mesh, 0, 0)
end

return rc
